-- Copyright (c) 2006 Frank Buss (fb@frank-buss.de)
-- See license.txt for license
--
-- DS2432 ROM id reader
--
-- Reads the 48 bit ROM id of a DS2432 interface, which is connected to a
-- one-wire pin at wire_io after startup and on reset, until the family
-- code and the CRC is valid. It asserts valid rom_id_o data with
-- data_valid_o = 1.
--
-- The timings are taken from http://www.maxim-ic.com/appnotes.cfm/an_pk/126
-- See http://en.wikipedia.org/wiki/1-Wire for more info on the 1-wire bus.

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity ds2432 is
  generic(system_speed: integer);
  port(
    clk_i: in std_logic;
    rst_i: in std_logic;
    rom_id_o: out unsigned(47 downto 0);
    data_valid_o: out std_logic;
    wire_io: inout std_logic);
end entity ds2432;

architecture rtl of ds2432 is
  constant us_delay: natural := system_speed / 1e6 - 1;
  signal us_delay_counter: natural range 0 to us_delay;

  signal wire_out: std_logic := '1';
  signal wire_in: std_logic;
  signal reset_result: std_logic;
  signal delay_counter: natural range 0 to 500;
  signal bit_counter: natural range 0 to 64;
  signal rom: unsigned(63 downto 0);
  signal crc: unsigned(7 downto 0);
  signal old_crc: unsigned(7 downto 0);

  constant a_delay: natural := 6;
  constant b_delay: natural := 64;
  constant c_delay: natural := 60;
  constant d_delay: natural := 10;
  constant e_delay: natural := 9;
  constant f_delay: natural := 55;
  constant g_delay: natural := 0;
  constant h_delay: natural := 480;
  constant i_delay: natural := 70;
  constant j_delay: natural := 410;
  
  type state_type is (
    reset1,
    reset2,
    reset3,
    reset4,
    reset5,
    write_command1,
    write_command2,
    write_command3,
    read_result1,
    read_result2,
    read_result3,
    wait_for_reset);

  signal state: state_type := reset1;
  
begin
  reader: process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_i = '1' then
        state <= reset1;
      else
        if us_delay_counter = 0 then
          us_delay_counter <= us_delay;
          case state is
            -- start reset
            when reset1 =>
              data_valid_o <= '0';
              rom_id_o <= (others => '0');
              wire_out <= '1';
              delay_counter <= g_delay;
              state <= reset2;
            when reset2 =>
              if delay_counter = 0 then
                wire_out <= '0';
                delay_counter <= h_delay;
                state <= reset3;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when reset3 =>
              if delay_counter = 0 then
                wire_out <= '1';
                delay_counter <= i_delay;
                state <= reset4;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when reset4 =>
              if delay_counter = 0 then
                reset_result <= wire_in;
                delay_counter <= j_delay;
                state <= reset5;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when reset5 =>
              if delay_counter = 0 then
                if reset_result = '0' then
                  -- device detected, start read ROM
                  rom(7 downto 0) <= x"33";
                  bit_counter <= 8;
                  state <= write_command1;
                else
                  -- not detected, try again
                  state <= reset1;
                end if;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when write_command1 =>
              if delay_counter = 0 then
                if bit_counter = 0 then
                  rom <= (others => '0');
                  bit_counter <= 64;
                  crc <= (others => '0');
                  state <= read_result1;
                else
                  -- send next bit
                  if rom(0) = '1' then
                    delay_counter <= a_delay;
                    state <= write_command2;
                  else
                    delay_counter <= c_delay;
                    state <= write_command3;
                  end if;
                  wire_out <= '0';
                  rom <= shift_right(rom, 1);
                  bit_counter <= bit_counter - 1;
                end if;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when write_command2 =>
              if delay_counter = 0 then
                wire_out <= '1';
                delay_counter <= b_delay;
                state <= write_command1;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when write_command3 =>
              if delay_counter = 0 then
                wire_out <= '1';
                delay_counter <= d_delay;
                state <= write_command1;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when read_result1 =>
              if delay_counter = 0 then
                if bit_counter = 0 then
                  if crc = 0 and rom(7 downto 0) = x"33" then
                    -- latch result
                    rom_id_o <= rom(55 downto 8);
                    data_valid_o <= '1';
                    state <= wait_for_reset;
                  else
                    -- else: crc error or wrong family code, read again
                    state <= reset1;
                  end if;
                else
                  -- read next bit
                  delay_counter <= a_delay;
                  state <= read_result2;
                  rom <= shift_right(rom, 1);
                  bit_counter <= bit_counter - 1;
                  wire_out <= '0';
                end if;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when read_result2 =>
              if delay_counter = 0 then
                delay_counter <= e_delay;
                wire_out <= '1';
                old_crc <= crc;
                state <= read_result3;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when read_result3 =>
              if delay_counter = 0 then
                rom(63) <= wire_in;
                delay_counter <= f_delay;
                crc <= old_crc(6) &
                       old_crc(5) &
                       (old_crc(4) xor (old_crc(7) xor wire_in)) &
                       (old_crc(3) xor (old_crc(7) xor wire_in)) &
                       old_crc(2) &
                       old_crc(1) &
                       old_crc(0) &
                       (old_crc(7) xor wire_in);
                state <= read_result1;
              else
                delay_counter <= delay_counter - 1;
              end if;
            when wait_for_reset =>
          end case;
        else
          us_delay_counter <= us_delay_counter - 1;
        end if;
      end if;
    end if;
  end process;
  
  wire_io <= wire_out when wire_out = '0' else 'Z';
  wire_in <= '1' when wire_io /= '0' else '0';

end architecture rtl;
